/*
 * Copyright (c) 1999 The Marcelo Ochoa (mochoa@exa.unicen.edu.ar).  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OWSKiller 
 *    Project for use in the OWSkiller servlet engine project
 *    (http://www.adela.sk/kuzela/OWSKiller)
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)."
 *
 * 4. The names "OWSKiller", "OWSKiller Servlet Engine" and 
 *    "OWSKiller Project" must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Products derived from this software may not be called "OWSKiller"
 *    nor "OWSKiller Server" appear in their names without 
 *    prior written permission of the OWSKiller Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OWSKiller 
 *    Project for use in the OWSKiller servlet engine project
 *    (http://www.adela.sk/kuzela/OWSKiller)
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)"
 *    
 * THIS SOFTWARE IS PROVIDED BY THE OWSKILLER PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the OWSKiller Group. For more information
 * on the OWSKiller Project and the OWSKiller Servlet Engine project,
 * please see <http://www.adela.sk/kuzela/OWSKiller/>
 * (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller).
 *
 */
package ows;

import java.util.*;
import javax.servlet.*; 
import javax.servlet.http.*;
import java.io.*;
import java.sql.*; //Package for JDBC classes

//Package for Oracle Extensions to JDBC
import oracle.sql.*;
import oracle.jdbc.driver.*;

public class ResourceManager implements Runnable
{ 
  static private int clients; // How many instances has this class

  static Stack cache;        // Store of free connections
  static int totalCount = 0; // Connection count gived to servlets
  static String JDBCDRIVER;
  static String DATABASE;    // DB driver connect string
  static String USER;        // DB username
  static String PASSWORD;    // DB password
  static int minConnections;     // Minimum of connections in Stack
  static ResourceManager instance;
  static boolean txEnable,dynLogin;
  static TxDicc dicc = null;
  static Hashtable activeTx;
  static Hashtable UserCache = null;
  static Hashtable PassCache = null;

  private Thread txThread = null;
  int counter = -1;
  String TxId;
  Connection connection;
  String user,password;

  static synchronized public ResourceManager getTxInstance(Properties props)
    throws  InstantiationException, IllegalAccessException, SQLException
  {
    if (dicc == null)
    {
        dicc = TxDicc.getInstance(props);
        activeTx =  new Hashtable();
	      UserCache =  new Hashtable();
	      PassCache = new Hashtable();
        instance = new ResourceManager();
        txEnable = "true".equalsIgnoreCase(props.getProperty("TX.Enable"));
        dynLogin = "true".equalsIgnoreCase(props.getProperty("Connection.DynamicLogin"));
        //System.out.println("Initializing Transaction manager...");
        JDBCDRIVER = props.getProperty("Connection.jdbcdriver");
        DATABASE = props.getProperty("Connection.database");
        USER = props.getProperty("Connection.dbusername");
        PASSWORD = props.getProperty("Connection.dbpassword");
        minConnections = (new Integer(props.getProperty("Connection.minconnections"))).intValue();

        cache = new Stack();

        // Load and register Oracle JDBC driver.
        try
        {
          //Load the Oracle JDBC Driver and register it.
          Class.forName(JDBCDRIVER).newInstance();
        }
        catch (Exception e)
        {
          System.out.println("Error loading JDBC driver");
        }
        openConnections(minConnections,USER,PASSWORD);
    }
    clients++;
    return instance;
  }

  public void run()
  {
    try {
        //System.out.println("Transaccion "+ TxId + " started...");
        while(counter-->=0) {
            txThread.sleep(1000);
            //System.out.println("Timeout in ="+counter);
        }
        try {
            //System.out.println("Transaccion "+ TxId + " aborted...");
            if (connection != null && !connection.isClosed()) {
                connection.rollback();
                connection.close();
            }
        } catch (Exception e) {
               System.out.println("Can't close connection DB in timeout trhead condition");
               e.printStackTrace();
        }
        // remove the transaccion from active TX
        activeTx.remove(TxId);
    } catch (Exception e) {
        System.out.println("Can't start the TX Thread or Thread interrup");
        e.printStackTrace();
    }
  }

  public ResourceManager() {
  }

  public ResourceManager(HttpServletRequest req,int timeOut, String usr, String pass)
      throws SQLException
{
      HttpSession ss = req.getSession(true);
      TxId = ss.getId();
      txThread = new Thread(this);
      activeTx.put(TxId,this);
      txThread.start();
      counter = timeOut;
      connection = DriverManager.getConnection(DATABASE,usr,pass);
      connection.setAutoCommit(false);
      //System.out.println("Connect without auto commit "+USER+"/"+PASSWORD);
  }

  /** Returns a connection from the cache. */
  public static synchronized Connection get(HttpServletRequest req, String usr, String pass)
    throws SQLException
  {
    //System.out.println("Get one connection From ResourceManager");
    if (txEnable) {
      // Process the session Enable behavior

      // Verify if the URL is part of a TxDef
      //System.out.println("Verify the URL ="+req.getRequestURI());
      TxDicc ctx = TxDicc.getTxDef(req.getRequestURI());
      if (ctx == null) {
        // if the URL is not part of any Transaction definition return a normal connection
        //System.out.println("Return normal connection");
        return get_connection(req , usr , pass);
      }
      // Get Session ID from de cookie session
      HttpSession ss = req.getSession(false);
      if (ss != null) {
        // if there are session id get the active connection from hastable
        ResourceManager rtmp = (ResourceManager) activeTx.get(ss.getId());
        if (rtmp == null) { // if there aren't Session Id in hashtable invalidate the request
          ss.invalidate(); // Invalidate the session ID, remove the cookie
          throw new SQLException("Invalid Session ID="+ss.getId());
        } else // return the connection if the execution path is valid
          switch(ctx.lastComp) {
              case TxDicc.CMMT: // Commit_URI
              case TxDicc.ROLL: // Rollback_URI
              case TxDicc.BLNG: // Belong_TO URI
                  if (rtmp.counter<0 || rtmp.connection.isClosed()) {
                     // if timeout end, remove and invalidate the transaccion
                     activeTx.remove(rtmp.TxId);
                     ss.invalidate(); // Invalidate the session, remove the cookie
                     throw new SQLException("Session aborted by timeout");
                  }
                  // Return the connection, the action take efect at release method.
                  return rtmp.connection;
              default: // There are session ID but is an ilegal action path close connection
                  // Stop the thread, remove from active TX and Invalidate the session ID
                  rtmp.txThread.stop();
		  if (!rtmp.connection.isClosed()) {
		    rtmp.connection.rollback();
		    rtmp.connection.close();
		  }
                  activeTx.remove(rtmp.TxId);
                  ss.invalidate(); // Invalidate the session ID, remove the cookie
                  throw new SQLException("Invalid action with Session ID="+ss.getId());
          } // end switch type of URI
      } else { // if there aren't ssesion id in the request verify the Begin of Tx
          if (ctx.lastComp==TxDicc.BGN) { // if is type Begin_URL
              ResourceManager txtmp = new ResourceManager(req,ctx.timeOut,usr,pass);
              return txtmp.connection;
          } else { // Return a normal connection
              return get_connection(req , usr , pass);
          }
      } // end if ss!=null
    } // end if txEnable
    return get_connection(req , usr , pass);
  }

  /** Returns a connection from the cache with user/password. */
  public static synchronized Connection get(HttpServletRequest req)
    throws SQLException
  {
    return get(req,USER,PASSWORD);
  }

  /** Returns true/false if the ResourceManager need the user/password
      at get time otherwise use the dbusername/dbpassword of the OWSKiller.cfg */
  public static synchronized boolean NeedLogin()
  {
    return dynLogin;
  }

  public static synchronized boolean isTxEnable(Connection conn)
    { // if have usr string in the UserCache then isn't Transaction Enable connection
    String usr = (String)UserCache.get(conn);
    return (usr==null);
  }

  public synchronized void release()
    throws SQLException
  {
    //System.out.println("release from ResourceManager");
    //System.out.println("release " + clients);
    if (--clients == 0) {
      while(!cache.empty())
        ((Connection)cache.pop()).close();
       UserCache.clear();
       PassCache.clear();
      //System.out.println("all released");
    }
    Enumeration e = activeTx.elements();
    while(e.hasMoreElements()) {
      ResourceManager txtmp = (ResourceManager) e.nextElement();
      if (txtmp.connection !=null && !txtmp.connection.isClosed()) {
         txtmp.connection.rollback();
         txtmp.connection.close();
      }
    }
  }

  /** Returns a connection to the cache. Or close connection. */
  public static synchronized void release(Connection connection, HttpServletRequest req)
    throws SQLException, NullPointerException
  {
    if (txEnable) {
      // Process the session Enable behavior
      // Verify if the URL is part of a TxDef
      //System.out.println("Verify the URL ="+req.getRequestURI());
      TxDicc ctx = TxDicc.getTxDef(req.getRequestURI());
      if (ctx == null) {
        // if the URL is not part of any Transaction definition release a normal connection
        //System.out.println("releasing normal connection from ResourceManager");
        release_connection(connection);
        return;
      }
      // Get Session ID from de cookie session
      HttpSession ss = req.getSession(false);
      if (ss != null) { // there are a cookie session ID verify the type of URI
        // if there are session id get the active connection from hastable
        ResourceManager rtmp = (ResourceManager) activeTx.get(ss.getId());
        if (rtmp == null) { // if there aren't Session Id in hastable invalidate the request
          ss.invalidate(); // Invalidate the session ID, remove the cookie
          throw new SQLException("Invalid Session ID="+ss.getId());
        } else // return the connection
          switch(ctx.lastComp) {
              case TxDicc.CMMT: // Commit transaccion
                  if (rtmp.counter<0 || rtmp.connection.isClosed()) {
                    // if timeout end, remove and invalidate the transaccion
                    activeTx.remove(rtmp.TxId);
                    ss.invalidate(); // Invalidate the session, remove the cookie
                    throw new SQLException("Session aborted by timeout");
                  }
                  //System.out.println("Commit the transaccion"+rtmp.TxId);
                  // Stop the Thread, remove the transaccion from active and commit
                  rtmp.txThread.stop();
                  rtmp.connection.commit();
                  rtmp.connection.close();
                  activeTx.remove(rtmp.TxId);
                  ss.invalidate(); // Invalidate the session, remove the cookie
                  return;
              case TxDicc.ROLL: // Rollback transaccion
                  if (rtmp.counter<0 || rtmp.connection.isClosed()) {
                    // if timeout end, remove and invalidate the transaccion
                    activeTx.remove(rtmp.TxId);
                    ss.invalidate(); // Invalidate the session, remove the cookie
                    throw new SQLException("Session aborted by timeout");
                  }
                  //System.out.println("Rollback the transaccion"+rtmp.TxId);
                  // Rollback, Stop the Thread and remove the transaccion from active
                  rtmp.txThread.stop();
                  rtmp.connection.rollback();
                  rtmp.connection.close();
                  activeTx.remove(rtmp.TxId);
                  ss.invalidate(); // Invalidate the session, remove the cookie
                  return;
              case TxDicc.BLNG: // Belong_TO URI, reset the counter
                  rtmp.counter=ctx.timeOut;
                  return;
              default: // nothing
                  return;
          } // end switch type of URI
      } else { // there aren't session ID release a normal connection
          //System.out.println("releasing normal connection from ResourceManager");
          release_connection(connection);
          return;
      } // end if ss!= null
    } // end if txEnable
    //System.out.println("releasing normal connection from ResourceManager");
    release_connection(connection);
  }
  /** Opens more connections. */
  static void openConnections(int count, String usr, String pass)
    throws SQLException
  {
    Connection ctmp;

    for(int i = 0; i < count; i++)
    {
      ctmp = DriverManager.getConnection(DATABASE,usr,pass);
      ctmp.setAutoCommit(true);
      UserCache.put(ctmp,usr);
      PassCache.put(ctmp,pass);
      //System.out.println("Connect N:" +i+" "+usr+"/"+pass);
      cache.push(ctmp);
    }
  }
  /** Returns a connection from the cache. */
  public static synchronized Connection get_connection(HttpServletRequest req, String usr, String pass)
    throws SQLException
  { Connection ctmp;
    if (cache.empty()) 
	    openConnections(1, usr, pass);
    totalCount++;
    //System.out.println("Get one connection");
    //System.out.println("Total Count ="+totalCount);
    ctmp = (Connection)cache.pop();
    String usrtmp = (String)UserCache.get(ctmp);
    String passtmp = (String)PassCache.get(ctmp);
    if (usrtmp.equalsIgnoreCase(usr) && passtmp.equalsIgnoreCase(pass)) {
	    // Return connection from Stack
	    //System.out.println("Get one connection from Stack =>"+usrtmp+"/"+passtmp+"@"+DATABASE);
	    return ctmp;
    } else {  // Reconnect with new usr/pass
	    UserCache.remove(ctmp);
	    PassCache.remove(ctmp);
	    ctmp.close();
	    ctmp = DriverManager.getConnection(DATABASE,usr,pass);
	    UserCache.put(ctmp,usr);
	    PassCache.put(ctmp,pass);
	    //System.out.println("Reconnect as =>"+usr+"/"+pass+"@"+DATABASE);
	    return ctmp;
    }
  }

  /** Returns a connection to the cache. Or close connection. */
  public static synchronized void release_connection(Connection connection)
    throws SQLException, NullPointerException
  {
    //System.out.println("releasing...");
    //System.out.println("Total Count ="+totalCount);
    if (totalCount-- < 0)
      throw new SQLException("Attempt to release more connections than allocated");

    if (connection == null || connection.isClosed())
      throw new NullPointerException("Attempt to release null or closed connection");

    if (cache.size() >= minConnections && cache.size() >= totalCount)
    {
      //System.out.println("removing : " + connection + " " + totalCount);
      UserCache.remove(connection);
      PassCache.remove(connection);
      connection.close();
      //System.out.println("closed");
    }
    else
    {
      connection.commit();
      cache.push(connection);
    }
    //System.out.println("Released.");
  }

}  // End of class
