/*
 * Copyright (c) 1999 The Marcelo Ochoa (mochoa@exa.unicen.edu.ar).  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OWSKiller 
 *    Project for use in the OWSkiller servlet engine project
 *    (http://www.adela.sk/kuzela/OWSKiller)
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)."
 *
 * 4. The names "OWSKiller", "OWSKiller Servlet Engine" and 
 *    "OWSKiller Project" must not be used to endorse or promote products 
 *    derived from this software without prior written permission.
 *
 * 5. Products derived from this software may not be called "OWSKiller"
 *    nor "OWSKiller Server" appear in their names without 
 *    prior written permission of the OWSKiller Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OWSKiller 
 *    Project for use in the OWSKiller servlet engine project
 *    (http://www.adela.sk/kuzela/OWSKiller)
 *    (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller)"
 *    
 * THIS SOFTWARE IS PROVIDED BY THE OWSKILLER PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the OWSKiller Group. For more information
 * on the OWSKiller Project and the OWSKiller Servlet Engine project,
 * please see <http://www.adela.sk/kuzela/OWSKiller/>
 * (http://www1.exa.unicen.edu.ar/cgi-bin/wrap/mochoa/OWSKiller).
 *
 */ 
package ows; 
 
import java.util.*; 
 
public class TxDicc  
{ 
  static private TxDicc instance = null; 
  static Properties properties = null; 
  static boolean txEnable; 
  static Hashtable dicc;
  String BeginURI; 
  String CommitURI; 
  String RollBackURI; 
  String []BelongTo; 
  int timeOut=0;
  String txName;
  int lastComp;
  static final int UNDF = 0;
  static final int BGN =  1;
  static final int CMMT = 2;
  static final int ROLL = 3;
  static final int BLNG = 4;
 
  static synchronized public  TxDicc getInstance(Properties props) 
  { 
    String txnames = props.getProperty("TX.names"); 
    if (instance == null) 
    { 
      //System.out.println("Initializing Transaction manager Dicc..."); 
      instance = new TxDicc(); 
      dicc = new Hashtable(); 
      properties = props; 
      txEnable = "true".equalsIgnoreCase(props.getProperty("TX.Enable")); 
 
      StringTokenizer st = new StringTokenizer(txnames," "); 
      while(st.hasMoreElements()) 
      { 
         String txdef = (String) st.nextElement(); 
         //System.out.println("\nAdd transaction "+txdef+" definition..."); 
         dicc.put( txdef ,new TxDicc(txdef)); 
      } 
    } 
    return instance; 
  } 
 
 
  public TxDicc() { 
  } 
 
  public TxDicc(String txdef) {
      int blngto = 0;
      txName = txdef;
      BeginURI = properties.getProperty(txdef+".BeginURI"); 
      CommitURI = properties.getProperty(txdef+".CommitURI"); 
      RollBackURI = properties.getProperty(txdef+".RollBackURI");
      timeOut = (new Integer(properties.getProperty(txdef+".TimeOut"))).intValue(); 
      //System.out.println("Add BeginURI ="+BeginURI+" definition..."); 
      //System.out.println("Add CommitURI ="+CommitURI+" definition..."); 
      //System.out.println("Add RollBackURI ="+RollBackURI+" definition..."); 
      //System.out.println("Add timeout ="+timeOut+" definition..."); 
      StringTokenizer st = new StringTokenizer(properties.getProperty(txdef+".BelongTo"),";");
      while(st.hasMoreElements()) {
         blngto++;
         st.nextElement();
      }
      st = new StringTokenizer(properties.getProperty(txdef+".BelongTo"),";");
      BelongTo = new String[blngto];
      for(int i=0;st.hasMoreElements();i++) {
         String s = (String)st.nextElement();
         BelongTo[i] = s.substring(0,s.indexOf("*"));
         //System.out.println("Add BelongTo ="+BelongTo[i]+" definition..."); 
      }
  } 
 
  /** Returns a Enumeration from the hash. */ 
  public static synchronized Enumeration getAll()
  {
      return dicc.elements();
  }
 
  public int getType(String URI)
  {
      if (URI.startsWith(BeginURI))
            return BGN;
      else if (URI.startsWith(CommitURI))
            return CMMT;
      else if (URI.startsWith(RollBackURI))
            return ROLL;       
      else for(int i=0;i<BelongTo.length;i++) {
            if (URI.startsWith(BelongTo[i]))
               return BLNG;
           }
      return UNDF;
  }

  public static synchronized TxDicc getTxDef(String URI)
  {
      Enumeration e = dicc.elements();
      while(e.hasMoreElements()) {
         TxDicc tx = (TxDicc) e.nextElement();
         if ((tx.lastComp=tx.getType(URI))!=UNDF)
            return tx;
      }
      return null;
  }

  public static synchronized String getTxName(String URI)
  {
      Enumeration e = dicc.elements();
      while(e.hasMoreElements()) {
         TxDicc tx = (TxDicc) e.nextElement();
         if ((tx.lastComp=tx.getType(URI))!=UNDF)
            return tx.txName;
      }
      return null;
  }
   
} 
